/*  $Id: App.java 1338 2012-08-25 17:38:30Z sdevaux $
 *
 * Tabs = 4 chars
 * copyright (C) 2010 Sebastien Devaux
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
 *
 *              Sebastien Devaux <sebastien.devaux@laposte.net>
 */
package net.eduvax.util;

import java.io.File;
import java.io.IOException;
import java.util.Date;
import java.util.Iterator;
import java.util.Hashtable;
import java.util.Locale;
import java.util.ResourceBundle;

/**
 * Currently running Application.
 * All main methods should start with App.init() or one subclass equivalent.
 * Should be the only real singleton of application. All other singletons are
 * relative to this one and may be reseted by init method (useful for unit testing
 * and test independence).
 * How to implements singletons using App class :
 * <pre>class Singleton {
 *  public Singleton get() {
 *      Singleton res=(Singleton)App.get().getSingleton(getClass());
 *      if (res==null) {
 *          res=new Singleton();
 *          CApp.get().setSingleton(res);
 *      }
 *		return res;
 *  }
 * }
 * </pre>
 * <b>warning:</b> in the example above, using getClass() to retrieve singleton
 * permits existence of several instance when effective type is different, ie
 * when subclassing singleton. When singleton is strict over all subclass, do not
 * use getClass() but Singleton.class (to be adapted upon the real Singleton types).
 * @djin.stereotype singleton
 * @version $Id: App.java 1338 2012-08-25 17:38:30Z sdevaux $
 * @author Sebastien Devaux
 */ 
public class App {

    ///////////////////////////////////////////////////////////////////
    // Init and configure services 
	/**
	 * Inits net.eduvax.util library.
	 * @param args initialisation arguments. standard use is to provide
	 * arguments from program's main method.
	 * Arguments matching the -Dname=value pattern are registered as 
	 * application attributes (as any attribute defined using 
	 * setAttribute() method).
	 */ 
    static public App init(String[] args) {
        if (_app!=null) {
            _app.log().warn("Application reset");
        }
        _app=new App(args);
        _app.setWorkingDir(".");
        return _app;
    }
	/**
	 * Inits net.eduvax.util library.
	 */ 
    static public App init() {
		return init(null);
    }
    /**
     * No public constructor. 
     * Use init() method to create App singleton.
     */ 
    private App(String args[]) {
        _attributes=new Hashtable<String,String>();
        if (args!=null) {
			setAttributesFromArgs(args);
        }
        Logger.init();        
        _logger=Logger.getSLogger(getClass().getName()); 
        _logger.info("Application start "+new Date());
        logPackageInfo(App.class);
        _errHandler=new AppLogErrHandler();
        setLocale(Locale.getDefault());
        _errHandlerFactory=new ErrHandlerFactory() {
            public ErrorHandler createErrHandler(int mode) {
                return new ConsoleErrHandler(mode);
            }
        };
    }
    /**
     * singleton getter.
     * May be null, to avoid this, main method MUST invoke init
     * before anything else.
     * @return App unique instance
     */
    public static App get() {
        return _app;
    } 
    /** singleton App instance */
    private static App _app;

    ///////////////////////////////////////////////////////////////////
    // singleton tools
    /**
     * Retrieve one singleton
     * @param type type of singleton
     */ 
    public Object getSingleton(@SuppressWarnings("rawtypes") Class type) {
        return _singletons.get(type);
    }
    /**
     * Set singleton instance for type.
     * @param type type of singleton
     * @param s new singleton instance
     */
    public void setSingleton(@SuppressWarnings("rawtypes") Class type,Object s) {
        Object exist=_singletons.get(type);
        if (exist==null) {
            _singletons.put(type,s);
        }
        else {
            _logger.warn("Trying to redefined singleton "+
                                    type+"["+exist+"] with: "+s);
        }
    }
    /**
     * Set singleton instance.
     * Type is automatically retrieve from effect given object's type.
     * Be care when type is subclassed, this permits creation of
     * one singleton for each subclass, so several instance of parent class...
     * To keep control of singleton type setting, 
     * use setSingleton(Class type, Object s).
     * @param s new singleton instance.
     */ 
    public void setSingleton(Object s) {
        setSingleton(s.getClass(),s);
    }
    /** singleton set */
    @SuppressWarnings("rawtypes")
	private Hashtable<Class,Object> _singletons=new Hashtable<Class,Object>();
    
    ///////////////////////////////////////////////////////////////////
    // Logging
    /** default logger */
    private SLogger _logger;
    /** get default logger */
    public SLogger log() {
        return _logger;
    }

	/**
	 * @return new logger
	 */ 
	public SLogger getLogger(String name) {
        return Logger.getSLogger(name); 
	}
    /** 
     * log package information
     * @param c reference class to use as package specifier
     */
    public void logPackageInfo(@SuppressWarnings("rawtypes") Class c) {
        Package p=c.getPackage();
        _logger.info(p.getName()+" V"+p.getImplementationVersion());
    } 
	/**
	 * Print trace.
	 * Print a debug log (using standard format). Message is preceded by
	 * TRACE[package.class.method():line_number]. Line number may not be
	 * available regarding the compilation debug options.
	 * @param msg message to be printed as trace.
	 */ 
    static public void TRACE(String msg) {
        StackTraceElement[] stack=Thread.currentThread().getStackTrace();
        String message;
        if (stack.length>2) {
            message="TRACE["+stack[2].getClassName()+"."+stack[2].getMethodName()
                    +"():"+stack[2].getLineNumber()+"]: "+msg;
        }
        else {
            message="TRACE:"+msg;
        }
        get().log().debug(message);
    }
    
    ///////////////////////////////////////////////////////////////////
    // Error Handling 
    /** application wide error tracer */
    private ErrorHandler _errHandler;
    /** application wide error handler factory */
    private ErrHandlerFactory _errHandlerFactory;
    /**
     * Get new handler.
     * Create new handler from application own errHandler factory.
     * Returned handler is chained to default error tracing handler.
     * @param mode error handling mode
     * @return error handler
     */ 
    public ErrorHandler getErrHandler(int mode) {
        return new ChainedErrHandler(_errHandlerFactory.createErrHandler(mode),_errHandler);
    }
    /**
     * Create handler from chaining given one to app error tracing handler.
     * @param h handler to chain
     */
    public ErrorHandler getErrHandler(ErrorHandler h) {
        return new ChainedErrHandler(h,_errHandler);
    }
    /**
     * Get app default error handler as dummy handler.
     * @return app default error handler.
     */
    public ErrorHandler getErrHandler() {
        return _errHandler;
    }
    /**
     * Change error handler factory.
     * Should be set once at application start.
     * @param factory new error handler factory
     */
    public void setErrHandlerFactory(ErrHandlerFactory factory) {
        _errHandlerFactory=factory;
    }

    ///////////////////////////////////////////////////////////////////
    // Localization
    /**
     * Change localization.
     * @param l new locale
     */ 
    public void setLocale(Locale l) {
        _locale=l;
        _logger.info("Set new locale : "+_locale);
    }
    /**
     * get current localization
     * @return current locale
     */ 
    public Locale getLocale() {
        return _locale;
    }
    /** Application wide localization */
    private Locale _locale;
    /**
     * Get resource bundle
     * @param name name of resource to load
     */
    public ResourceBundle getResource(String name) {
        return ResourceBundle.getBundle(name,_locale);
    }

    ///////////////////////////////////////////////////////////////////
    // File & working dir utilities
    private File _workingDir;
    /**
     * Change app pseudo working directory
     * @param dirpath new path
     */
    public void setWorkingDir(String dirpath) {
        try {
            _workingDir=(new File(dirpath)).getCanonicalFile();
            _logger.debug("Working dir changed: "+_workingDir);
        } 
        catch (IOException ex) {
            _logger.warn("Can't change working directory: "+ex+
                ". Keeping current one: "+_workingDir);
        }
    }
    /**
     * Get current pseudo working directory
     * @return app working directory
     */
    public File getWorkingDir() {
        return _workingDir;
    }
    /**
     * Get File relative to app pseudo working directory.
     * Return path is relative to app working directory only when
     * given path is not absolute.
     */
    public File getFile(String path) {
        File f=new File(path);
        if (!f.isAbsolute()) {
            f=new File(_workingDir,path);
        }
        return f;
    }

    ///////////////////////////////////////////////////////////////////
    // Misc utilities 
    /**
     * Exit application.
     * No System.exit() should be invoke from elsewhere than here...
     * @param cr exit code
     */
    public void exit(int cr) {
		Iterator<EThread> iTh=EThread.getActiveThreads();
		while(iTh.hasNext()) {
			EThread th=iTh.next();
			if (!th.isCurrentThread())
			{
				log().warn("Thread "+th.getName()+" still active, Aborting");
				th.interrupt();
			}
		}
		_singletons=null;
        System.exit(cr);
    }
    /**
	 * Exit application.
	 * Alias to exit(0).
	 */ 
    public void exit() {
		exit(0);
    }
    /**
     * Generate a time stamp.
     * Default implementation is current time in milliseconds.
     * @return time stamp.
     */
    static public long getTimestamp() {
		return System.currentTimeMillis();
    }
    
    //////////////////////////////////////////////////////////////
	/** application wide variables. Environment variable complement. */
    private Hashtable <String,String> _attributes;
    /**
	 * Defines new application attribute.
	 * Value is overwrite when attribute is alreadey defined.
	 * @param name attribute's name
	 * @param val attribute's value
	 */
    public void setAttribute(String name,String val) {
		_attributes.put(name, val);
    }
    /**
	 * set attribute only if not already set.
	 * @param name new attribute's name.
	 * @param val new attribute's value.
	 */
    public void setUnsetAttribute(String name,String val) {
		if (_attributes.get(name)==null) {
			_attributes.put(name, val);
		}
    }

    /**
	 * Gets an attribute or environment variabel value. 
	 * Attribute seach order is: 1st system property (eg env var), 2nd
	 * application attribute. The main consequence is, it is not possible
	 * to override system attribute with application attributes.
	 * An attribute value may contain in its definition a ${attr_name} pattern.
	 * When the corresponding attribute name is defines, the pattern is 
	 * expended with the attribute's value in the resulting string. Expension
	 * occurs during getAttribute call, so subsequent calls may return 
	 * different string if some inclosed attributes values are changing.
	 * @see expandFromAttributes
	 * @param name attribute's name
	 * @return attributes value (other attributes reference expanded).
	 */
    public String getAttribute(String name) {
		String res=System.getenv(name);
		if (res==null) {
			res=System.getProperty(name);
		}
		if (res==null) {
			res=_attributes.get(name);
		}
		if (res!=null) {
			res=expandFromAttributes(res);
		}
		return res;
    }
    private final static int EES_TEXT=0;
    private final static int EES_VB=1;
    private final static int EES_VNAME=2;
    private final static char EE_VB1='$';
    private final static char EE_VB2='{';
    private final static char EE_VE='}';
	/**
	 * Expand attribute reference in a String.
	 * Provide new string by replacing ${attr_name} patterns with
	 * attributes value. When a pattern is refering an unknown attribute,
	 * the empy string is used for its value (then the effect in reterned 
	 * string is the pattern appears to have been removed).
	 * @warning no control is made on inclose attributes reference. Infinite 
	 * loop will occure in case of recusive references.
	 * @return expanded String. May be equal to input string if no reference
	 * to a an attribute is present in input string.
	 */ 
    public String expandFromAttributes(String str) {
		StringBuffer res=new StringBuffer(str.length());
		StringBuffer var=new StringBuffer();
		int state=EES_TEXT;
		for (int i=0;i<str.length();i++) {
			char c=str.charAt(i);
			switch (state) {
				case EES_VB:
					if (c==EE_VB2) {
						state=EES_VNAME;
						var=new StringBuffer();
					}
					else {
						state=EES_TEXT;
						res.append(EE_VB1);
					}
					break;
				case EES_VNAME:
					if (c==EE_VE) {
						String val=getAttribute(var.toString());
						res.append(val=="null"?"":val);
						state=EES_TEXT;
					}
					else {
						var.append(c);
					}
					break;
				default:
					if (c==EE_VB1) {
						state=EES_VB;
					}
					else {
						res.append(c);
					}
					break;
			}
		}
		return res.toString();
    }
    /**
	 * Defines attributes from an arguments list.
	 * Use all -Dname=value found in list as an attribute definition.
	 * @param args arguments to parse.
	 */ 
    private void setAttributesFromArgs(String[] args) {
		for (int i=0;i<args.length;i++) {
			if (args[i].startsWith("-D")) {
				int pos=args[i].indexOf('=');
				if (pos>0)
				{
					String pname=args[i].substring(2,pos-2);
					String pval=args[i].substring(pos+1);
					setAttribute(pname,pval);
				}
			}
		}
    }
 }
